#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "glut.h"
#include <string.h>
#include "bmp.h"
#include "mesh.h"

#define TITLE "IBMR Starter Code"
#define WIDTH 512
#define HEIGHT 512

#define SCREENSHOT "screenshot.bmp" // filename for screenshot
float x_pos= -0.9, y_pos=0.9; 	 // location to display the text messages
char message[80]; 	// message to display
int meshW=100,meshH=100; // mesh width and height
static GLuint texName;	 // this references the texture map
mesh myMesh;	         // contains the mesh
Mat3 H;	                 //  the Projective transformation matrix
MODE mesh_display_mode=FILL;
bool enableTexture=true;

// outputs text pointed by string on OpenGL screen at location x,y
void OutputString (  float x,  float y,  char *string )
{
	int length;

	length = strlen ( string );
	glRasterPos2f ( x, y );
	
	for (int i = 0; i < length; i++ )
		glutBitmapCharacter ( GLUT_BITMAP_9_BY_15, string[i] );
}

// Saves whatever is currently in the front frame buffer to the file specified
// Returns -1 on error, 0 on success;
int save_screenshot(char *fname)
{
	int err;
	unsigned char *image;

	image = (unsigned char *) malloc(WIDTH * HEIGHT * 4); // RGBA

	glPixelStorei(GL_PACK_ALIGNMENT, 1); // set correct alignment
	glReadBuffer(GL_FRONT); // we want to read the back buffer

	// read the entire display window
	glReadPixels(0, 0, WIDTH, HEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, image);

	// Now store it to a file
	err = save_BMP(fname, image, WIDTH, HEIGHT);

	free(image); // we do not want a memory leak!

	return err;
}


// glut display function. Called each time screen is to be redrawed. Can be forced
// by calling glutPostRedisplay().
// This is where all the drawing related stuff takes place.

void display()
{
	// Everything we do here gets done in the back buffer. The buffers are swaped
	// after we are done with the changes.

	// Clear the frame-buffer
	glClear(GL_COLOR_BUFFER_BIT );
	if(enableTexture) glEnable(GL_TEXTURE_2D);	// May not be used in Linux
	myMesh.display(); // display  the mesh
	if(enableTexture) glDisable(GL_TEXTURE_2D); // May not be used in Linux
	OutputString(x_pos, y_pos,message);
	glFlush();	// refresh the buffer

	// Swap the frame buffer. Make the back buffer the new front buffer and
	// vice-versa. To avoid flicker we use two buffers. While we display one,
	// we make changes to the second hidden buffer.
	glutSwapBuffers();
}


// glut reshape function. Called when the window is resized. Also called when the window
// is created. Initialize the Projection matrix here.
void reshape(int w, int h)
{
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);

	// Set the projection matrix
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(-1.0, 1.0, -1.0, 1.0);

	// Initialize the modelview matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

// glut keyboard function.
void keyboard(unsigned char key, int x, int y)
{
	switch(key) {
		case 'S':
		case 's':
			// Save whatever is currently on screen to a file
			if(save_screenshot(SCREENSHOT))
				sprintf(message,"Could not save to file %s.\n", SCREENSHOT);
			else
				sprintf(message,"Current display saved to file %s.\n", SCREENSHOT);
			break;

		case 27: // Esc
		case 'Q':
		case 'q':
			exit(0); // Quit from application
			break;
		case 'h':
		case 'H': 
			// apply the transform to the mesh
			myMesh.ApplyTransform(H);
			sprintf(message,"H key pressed: Applied transformation");
			break;
		case 'm':
		case 'M':
			// toggle between lines, points and fill mode
			switch(mesh_display_mode)
			{
			case LINE: 
				mesh_display_mode = POINT;
				sprintf(message,"Mesh Display Mode: POINT");
				break;
			case POINT: 
				mesh_display_mode = FILL;
				sprintf(message,"Mesh Display Mode: FILL");
				break;
			case FILL: 
				mesh_display_mode = LINE;
				sprintf(message,"Mesh Display Mode: LINE");
				break;
			};
			break;
		case 'r':
		case 'R':
			// reset
			myMesh.reset();
			sprintf(message,"resetting the mesh");
			break;
		case 't':
		case 'T':
			// toggle texture
			enableTexture = !enableTexture;
			if(enableTexture)
			{
				sprintf(message,"Texture Mapping turned ON");
				glEnable(GL_TEXTURE_2D);
			}
			else
			{
				sprintf(message,"Texture Mapping turned OFF");
				glDisable(GL_TEXTURE_2D);
			}
			break;	
	}

	// We might have changed something. Force a re-display
	glutPostRedisplay();
}

// glut special key handler
void special(int key, int x, int y)
{
	// As a demonstration, we change two global variables based on
	// the key press. These variables are used to translate the text
	// we have drawn.
	switch(key)	{
		case GLUT_KEY_UP:
			y_pos+=0.01;
			break;

		case GLUT_KEY_DOWN:
			y_pos-=0.01;
			break;
		case GLUT_KEY_LEFT:
			x_pos-=0.01;
			break;
		case GLUT_KEY_RIGHT:
			x_pos+=0.01;
			break;
		default:
			break;
	}

	// We might have changed something. Force a re-display
	glutPostRedisplay();
}

// glut idle function. Called at every 'clock tick'
void idle(void)
{
	// Put anything that needs to be updated all the time here.
	// Any sort of animation would require you to use this fundtion.
	// Remember to call glutPostRedisplay() once you've made the changes
	// to force the screen to update

	// We might have changed something. Force a re-display
	glutPostRedisplay();
}

// Initialize all opengl related stuff
void gl_init()
{
	glClearColor(0.0, 0.0, 0.0, 0.0); // Set clear color
}

// Initialize all glut related stuff, including callback functions.
void glut_init(int *argc, char **argv)
{
	// Call glut's initilization
	glutInit(argc, argv);

	// double buffer, RGB color model, depth buffer
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB); // | GLUT_DEPTH);

	// Now create a window of the required size
	glutInitWindowSize(WIDTH, HEIGHT);
	glutInitWindowPosition(0, 0);
	glutCreateWindow(TITLE);

	// glut Callback functions
	glutDisplayFunc(display); // display
	glutKeyboardFunc(keyboard); // keyboard
	glutSpecialFunc(special); // special keys
	glutReshapeFunc(reshape); // reshape
	glutIdleFunc(idle); // idle


	// These are callbacks for the mouse. Implement yourself if needed
	glutMouseFunc(NULL);
	glutMotionFunc(NULL);
}

void LoadTexture()
{
  // Try loading a BMP image file - for illustration
	int w, h;
	unsigned char *img;

	int success = load_padded_BMP("image.bmp", &img, &w, &h);
	if(success!=0)
	{
		printf("could not load bmp file image.bmp\n");
		enableTexture = false;
		return;
	}
		
	printf("Loaded BMP file. Width = %d, Height = %d\n", w, h);
	
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glClear(GL_COLOR_BUFFER_BIT);

	// generate 1 texture
	glGenTextures(1, &texName);


	// initialize the texture
	glBindTexture(GL_TEXTURE_2D, texName);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, img);

 
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glEnable(GL_TEXTURE_2D); // default texture enable
}

void init()
{
	myMesh.Initialize(meshW,meshH);	// build the mesh
	LoadTexture();	
	//*****************************************
	// SET YOUR H MATRIX VALUES HERE!!!
	Mat3 S;

	//*****************************************
	// SET YOUR H MATRIX VALUES HERE!!!
	//*****************************************
	S.makeScale(0.5,0.75);
	H.ident();
	H.mpy(H,S);
  	S.makeRot(45.0*M_PI/180.0);
  	H.mpy(H,S);
	H.prnt();
	sprintf(message,"Welcome to IBMR. Please see console for options");
	cout << "List of keys:\nt : Toggle Texture ON/OFF\nm : Toggle display mode between POINT/LINE/FILL\nh : Apply Transformation recursively\nr : Reset mesh to start\nq,Esc : Quit\nArrow keys: change text location\n";
}

int main(int argc, char** argv)
 {
	// First initialize glut
	glut_init(&argc, argv);
	// Now initialize OpenGL stuff
	gl_init();
	// Now initialize program specific stuff
	init();
	// Finally enter the glut infinite loop 
	glutMainLoop();

	return 0;
}
